# -*- coding: utf-8 -*-#
# Python script that uses the consistency excel file, cytometer data and quantiFlash to evaluate date, it will create SNR and DNR plots for all selected detectors

# @author: Sjoerd Visser, APE GmbH, Berlin
# @author: Konrad von Volkmann, APE GmbH, Berlin
# @author: Stephan Benke, Cytometry Facility, University Zurich
# @author: Toralf Kaiser, DRFZ, Berlin
# @author: Sebastian Falkowski, APE GmbH, Berlin
# $Revision: 139 $

SVNRevision = '$Revision: 139 $'

# -------------------------------------
# Don't change or delete this section
# -------------------------------------
# @inistart

# [script]
# name = 3. Eval. SNR DNR

# [param1]
# name = Directory
# label = Data Directory
# value = C:\Temp\Messungen
# global_var = DataDir
# mandatory = true
# datatype = directory

# @iniend

# -------------------------------------------------------------------------------
# Preperations
# -------------------------------------------------------------------------------

import fcsparser
import numpy as np
import pandas as pd
import argparse
import configparser
import glob, os, fnmatch, re
import re
import scipy.stats as sp

# Matplotlib packages
# import matplotlib.pylab as p
# import matplotlib.gridspec as gridspec
# import matplotlib.patches as patches
# plt.ioff()

from ape.gui import Control

# Excel packages?
import openpyxl
from openpyxl import Workbook
from openpyxl import load_workbook
from openpyxl import formatting, styles
from openpyxl.chart import  Reference, Series, ScatterChart
from openpyxl.chart.shapes import GraphicalProperties
from openpyxl.styles import  Border, Side, Font, Alignment
from openpyxl.utils import get_column_letter

import warnings
warnings.filterwarnings("ignore")

# -------------------------------------
# Script version
script_version = '1.0.130'
# -------------------------------------


# ----------------------
# Functions
# ----------------------

def load_file(file,V):
    meta, df_data = fcsparser.parse(file, reformat_meta=True)
    return df_data

def ramp_evaluation(df_ramp):

    # Seperate the intensities via a jump in time diff > 5
    timing = df_ramp.loc[:,'Time'].values
    df_ramp.drop('Time', axis=1, inplace=True)
    diff = timing - np.roll(timing,1)

    sections = np.where(diff>5)[0]
    # sections = sections[sections>11000]
    section0 = 0
    for nr,section1 in enumerate(sections):
        df_temp = df_ramp.loc[section0:section1,:]
        section0 = section1

        gmean = pd.Series(sp.stats.gmean(df_temp,axis=0), index=df_ramp.columns)

        # gstd_factor = np.exp(np.sqrt(np.sum(np.log(df_temp/gmean)**2,axis=0)/np.size(df_temp,axis=0)))
        # gstd = (gmean*gstd_factor-gmean/gstd_factor)/2
        # gstd = pd.Series(gstd)
        gmean = df_temp.mean()
        gstd = df_temp.std()
        CV = gstd/gmean
        median = df_temp.median()

        npe = gmean**2/gstd**2
        K_factor = npe/gmean

        if nr == 0:
            #initialize dataframes
            df_gmean = pd.DataFrame(gmean)
            df_gstd = pd.DataFrame(gstd)
            df_CV = pd.DataFrame(CV)
            df_median = pd.DataFrame(median)

            df_npe = pd.DataFrame(npe)
            df_K_factor = pd.DataFrame(K_factor)

        else:
            df_gmean = pd.concat([df_gmean, gmean], axis=1)
            df_gstd = pd.concat([df_gstd, gstd], axis=1)
            df_CV = pd.concat([df_CV, CV], axis=1)
            df_median = pd.concat([df_median, median], axis=1)

            df_npe = pd.concat([df_npe, npe], axis=1)
            df_K_factor = pd.concat([df_K_factor, K_factor], axis=1)

    # Set column names and transpose
    NN=df_gmean.shape[1]
    df_gmean.columns = intensities[:NN]
    df_gmean.index = df_ramp.columns
    df_gmean = df_gmean.T

    df_gstd.columns = intensities[:NN]
    df_gstd.index = df_ramp.columns
    df_gstd = df_gstd.T

    df_CV.columns = intensities[:NN]
    df_CV.index = df_ramp.columns
    df_CV = df_CV.T

    df_median.columns = intensities[:NN]
    df_median.index = df_ramp.columns
    df_median = df_median.T

    df_npe.columns = intensities[:NN]
    df_npe.index = df_ramp.columns
    df_npe = df_npe.T

    df_K_factor.columns = intensities[:NN]
    df_K_factor.index = df_ramp.columns
    df_K_factor = df_K_factor.T

    cond =  ((df_gmean>100)&(df_gmean<1e4))
    cond2 = cond*1
    cond3 = cond2.T*intensities[:NN].T
    cond3[cond3==0] = np.nan

    data = np.zeros([4,cond.shape[1]])
    df_K_factor_info = pd.DataFrame(data,index=['points taken','max int','min int','mean'],columns=cond.columns)
    df_K_factor_info.loc['points taken',:] = cond2.sum()
    df_K_factor_info.loc['min int',:] = cond3.T.max() # switch around because large values are small intensities
    df_K_factor_info.loc['max int',:] = cond3.T.min()
    df_K_factor_info.loc['mean'] = df_K_factor.where(cond).mean()
    df_K_factor_mean = df_K_factor.where(cond).mean()

    return df_gmean, df_gstd, df_CV, df_median, df_npe, df_K_factor, df_K_factor_mean, df_K_factor_info

def lib_evaluation(df_lib):
    df_lib.drop('Time', axis=1, inplace=True)
    df_prctile_lib = df_lib.quantile(0.99, axis=0)

    return df_prctile_lib

def snr_dnr(df_npe_100npe, df_Npe_lib, df_K_factor_mean, detector_select, resolution):

    if detector_select == 'heigth':
        df_SNR = 20.0*np.log10(np.abs(df_npe_100npe/df_Npe_lib))
        df_DNR = 20.0*np.log10(np.abs(df_K_factor_mean/df_Npe_lib)*resolution)

    if detector_select == 'area':
        df_SNR = 10.0*np.log10(np.abs(df_npe_100npe/df_Npe_lib))
        df_DNR = 10.0*np.log10(np.abs(df_K_factor_mean/df_Npe_lib)*resolution)

    if detector_select == 'both':
        df_SNR = 10.0*np.log10(np.abs(df_npe_100npe/df_Npe_lib))
        df_DNR = 10.0*np.log10(np.abs(df_K_factor_mean/df_Npe_lib)*resolution)

        for detector in df_SNR.index.values:
            if detector[-1]=='H':
                df_SNR.loc[detector,:] *= 2
                df_DNR.loc[detector,:] *= 2
    return df_SNR, df_DNR

def optimum_voltage(df_SNR_total, df_DNR_total):
    df_range = df_SNR_total + df_DNR_total

    V_opt = df_range.idxmax(axis=1)
    V_max = V_opt.copy()
    V_min = V_opt.copy()

    for detector in df_range.index:
        temp = df_range.loc[detector,:]
        cond = temp < temp.max()-3
        temp = temp.where(cond)
        if not(np.isnan(temp)).all():
            V_max[detector] = temp.loc[V_opt[detector]:].idxmax()
            V_min[detector] = temp.loc[:V_opt[detector]].idxmax()

    df_Vopt = pd.concat([V_min, V_opt, V_max], axis=1)
    df_Vopt.columns = ['Vmin','Vopt','Vmax']

    return df_Vopt

def init_excel(fname, sheets, titels):
    wb = Workbook()

    for nr, name in enumerate(sheets):
        wb.create_sheet(name)
        ws = wb[name]
        ws['A1'] = titels[nr]

        # Title banner from A1:Z1
        cell_range = ['%s1' % get_column_letter(nr) for nr in range(1,27)]
        # Banner style
        blue_fill = styles.PatternFill("solid", fgColor="4398ED")
        White_color_font = 'FFFFFF'
        White_font = styles.Font(size=28, bold=True, color=White_color_font)

        for i in cell_range:
            cell = ws[i]

            cell.fill = blue_fill
            cell.font = White_font

    wb.remove(wb['Sheet'])
    wb.save(fname)

def export_results(df_results,
                   df_SNR,
                   df_DNR,
                   df_npe_lib,
                   df_lib_prctile,
                   df_npe_ref,
                   df_gmean_ref,
                   df_median_ref,
                   df_gstd_ref,
                   df_CV_ref,
                   df_K_factor_mean,
                   intensity_npe_100,
                   data_folder):

    # Initialization of excel
    fname = os.path.join(directory, 'Results.xlsx')
    sheets = ['Summary']

    detectors = df_SNR.index.values
    for detector in detectors:
        detector_clean = re.sub('[^A-Za-z0-9]+', ' ', detector)
        sheets = np.append(sheets, detector_clean)

    titles = ['%s' %string for string in sheets]
    init_excel(fname, sheets, titles)

    wb = load_workbook(fname)
    writer = pd.ExcelWriter(fname, engine='openpyxl')
    writer.book = wb
    writer.sheets = dict((ws.title, ws) for ws in wb.worksheets)

    offset=4
    # Summary
    df_results.to_excel(writer, sheet_name=sheets[0], startrow=offset)
    ws = wb[sheets[0]]
    ws.column_dimensions['A'].width = 40
    sheetnr = 1

    # titles
    White_font = styles.Font(size=16, bold=True)

    # Fill Detector sheets
    for nr, sheet in enumerate(sheets[sheetnr:]):
        offset=8
        # Detector specific dataframe
        # Signal

        df_detector1 = pd.concat([df_gmean_ref.iloc[nr,:],
                                 df_median_ref.iloc[nr,:],
                                 df_gstd_ref.iloc[nr,:],
                                 df_CV_ref.iloc[nr,:],
                                 df_K_factor_ref.iloc[nr,:],
                                 df_npe_ref.iloc[nr,:]], axis=1)
        df_detector1.columns = ['geo mean \t [ch]','median \t [ch]','geo std \t [ch]','CV \t [%]','K-factor \t [Npe/ch]','Npe']
        df_detector1.T.to_excel(writer, sheet_name=sheet, startrow=offset)
        offset+=9

        # Background
        df_detector2 = pd.concat([df_lib_prctile.loc[detectors[nr]],
                                  df_K_factor_mean.loc[detectors[nr]],
                                  df_npe_lib.loc[detectors[nr]]], axis=1)
        df_detector2.columns = ['99% point','K-factor','99% point [Npe]']
        df_detector2.T.to_excel(writer, sheet_name=sheet, startrow=offset)
        offset+=6

        # Results
        df_detector3 = pd.concat([df_SNR.loc[detectors[nr]], df_DNR.loc[detectors[nr]]], axis=1)
        df_detector3.columns = ['SNR','DNR']
        df_detector3.T.to_excel(writer, sheet_name=sheet, startrow=offset)



        ws = wb[sheet]
        ws.column_dimensions['A'].width = 20

        ws['A2'] = detectors[nr]
        ws['A8']='Signal'
        cell = ws['A8']
        cell.font = White_font

        ws['A4']= 'The reference signal (100npe)'
        cell = ws['A4']
        cell.font = White_font
        ws['A5']= 'Quantiflash intensity:'
        ws['D5']= str(intensity_npe_100[nr])
        ws['E5']= 'dB'
        ws['A6']= 'Reference Voltage:'
        ws['D6']= str(V_ref)
        ws['E6']= 'V'

        ws['A17']='Background'
        cell = ws['A17']
        cell.font = White_font

        ws['A23']='Results'
        cell = ws['A23']
        cell.font = White_font
    wb.save(fname)
    wb.close()



    # Excel graphs
    wb = load_workbook(fname)
    for nr, sheet in enumerate(sheets[sheetnr:]):
        ws = wb[sheet]
        c1 = ScatterChart()
        c1.title = str(ws['A2'].value)
        c1.y_axis.title = 'Signal [dB]'
        c1.x_axis.title = 'Voltages [V]'
        c1.x_axis.scaling.min = np.min(voltages)-50
        c1.x_axis.scaling.max = np.max(voltages)+50
        c1.y_axis.scaling.min = 0
        c1.y_axis.scaling.max = 90
        c1.height = 12
        c1.width = 20

        VOLTAGE = Reference(ws, min_col=1, max_col=1+np.size(voltages), min_row=offset+1, max_row=offset+1)

        SNR = Reference(ws, min_col=1, max_col=1+np.size(voltages), min_row=offset+2, max_row=offset+2)
        DNR = Reference(ws, min_col=1, max_col=1+np.size(voltages), min_row=offset+3, max_row=offset+3)

        s1 = Series(DNR, xvalues=VOLTAGE, title="DNR")
        s2 = Series(SNR, xvalues=VOLTAGE, title="SNR")

        color1='e66101'
        s1.graphicalProperties.line.width = 50000
        s1.graphicalProperties.line.solidFill = color1          # Line Fill
        s1.marker.graphicalProperties.solidFill = color1        # Marker filling
        s1.marker.graphicalProperties.line.solidFill = color1   # Marker outline
        s1.marker.symbol = "circle"
        s1.marker.size = 10

        color2='fdb863'
        s2.graphicalProperties.line.width = 50000
        s2.graphicalProperties.line.solidFill = color2          # Line Fill
        s2.marker.graphicalProperties.solidFill = color2        # Marker filling
        s2.marker.graphicalProperties.line.solidFill = color2   # Marker outline
        s2.graphicalProperties.line.dashStyle = "sysDot"        # Dotted line
        s2.marker.symbol = "circle"
        s2.marker.size = 10


        c1.append(s1)
        c1.append(s2)

        ws.add_chart(c1, "P5")

        wb.save(fname)
    wb.close()


# ----------------------
# Preperations
# ----------------------
argparser = argparse.ArgumentParser(description='Description')
argparser.add_argument('--Directory', metavar='Directory', type=str)
args = argparser.parse_args()
data_folder = args.Directory

gui = Control()
log_thresh_for_nan = 1e-3
npe_ref = 100

# ======================
# Evaluation of SNR and DNR
# ======================
global voltages, detectors, intensities, V_ref

print('------------------------------------')
print('Script: 3: Evaluation data to SNR and DNR')
print('------------------------------------')
print('\n')

# ----------------------
# Check if folders and data structure
# ----------------------
ramp_folder = os.path.join(data_folder, 'ramp')
# ramp_filter = os.path.join(ramp_folder, '*.fcs')
lib_folder = os.path.join(data_folder,'LIB')
# lib_filter = os.path.join(lib_folder,'*.fcs')

if not os.path.exists(data_folder):
    gui.error('Data folder not found.\nPlease check the "Evaluation Directory"')
else:
    # Check for Evaluation folder, if not -> create
    directory = os.path.join(data_folder,'Evaluation')
    if not os.path.exists(directory):
        gui.error('Evaluation folder not found.\nPlease run the script "2. Data Consistency" first')
if not os.path.exists(os.path.join(data_folder,'ramp')):
    gui.error('Ramp data not found.\nPlease check the correct data structure')
else:
    ramp_files = fnmatch.filter(os.listdir(ramp_folder), '*.fcs')
    if np.size(ramp_files) == 0:
        gui.error('Ramp data not found.\nPlease check the correct data structure')
if not os.path.exists(os.path.join(data_folder,'LIB')):
    gui.error('LIB data not found.\nPlease check the correct data structure')
else:
    lib_files = fnmatch.filter(os.listdir(lib_folder), '*.fcs')
    if np.size(lib_files) == 0:
        gui.error('LIB data not found.\nPlease check the correct data structure')


# ----------------------
# Load qF settings file (cfg) from main data folder
# ----------------------
cfgFilter = os.path.join(data_folder,'quantiFlash_script_settings*.cfg')
cfgFiles = glob.glob(cfgFilter)
# cfgFile = os.path.join(data_folder,'quantiFlash_script_settings.cfg')
if os.path.isfile(cfgFiles[0]):
    Config = configparser.ConfigParser()
    Config.read(cfgFiles[0])

    # Device settings
    device = str(Config.get('Main','device'))
    port_id = int(Config.get('Main','port'))

    # QuantiFlash settings
    startInt = float(Config.get('quantiFlash','startInt'))
    stopInt = float(Config.get('quantiFlash','stopInt'))
    stepInt = float(Config.get('quantiFlash','stepInt'))
    triggerInt = float(Config.get('quantiFlash','triggerInt'))
    countRate = float(Config.get('quantiFlash','countRate'))
    nSamples = float(Config.get('quantiFlash','nSamples'))
    time_to_sleep = float(nSamples/countRate)-0.1

    # Evaluation settings
    detector_select = str(Config.get('Evaluation','detectors'))
    resolution = int(Config.get('Evaluation','resolution'))
    V_ref = int(Config.get('Evaluation','v_ref'))

    # Output settings
    raw_SD = bool(Config.get('Output','raw_SD'))
    raw_mean = bool(Config.get('Output','raw_mean'))
    raw_median = bool(Config.get('Output','raw_median'))
    raw_LIB = True
    K_factor = bool(Config.get('Output','K_factor'))
    result = bool(Config.get('Output','result'))
else:
    if gui.confirm("No quantiFlash settings file found.\n Continue with default parameters?") != 'OK':
        sys.exit(1)
    else:
        startInt = -94
        stopInt = 0
        stepInt = 2
        triggerInt = 50
        countRate = 4000.0
        nSamples = 10000.0
        time_to_sleep = float(nSamples/countRate)

        # Evaluation settings
        detector_select = 'height'
        resolution = 262144
        V_ref = 550

        # Output settings
        raw_SD = True
        raw_mean = True
        raw_median = False
        raw_LIB = True
        K_factor = True
        result = True
        
intensities = np.arange(startInt, stopInt+stepInt, stepInt)
if intensities[-1]>0:
    intensities = np.delete(intensities,-1)

# ----------------------
# Load Data
# ----------------------
print('\n --- Loading files --- \n')

# ----------------------
# Load ConsistencyCheck.xlsx data
# ----------------------
file_name = os.path.join(directory,'ConsistencyCheck.xlsx')
df_error_log = pd.read_excel(file_name,'Error log',header=0)
df_error_log.rename(columns={'Unnamed: 0':'voltage'}, inplace=True )
df_no_error = df_error_log[(df_error_log['ramp_files']!=0)&(df_error_log['lib_files']!=0)&(df_error_log['ramp_complete']!=0)]

ramp_files = np.array(df_no_error['ramp_files'])
lib_files = np.array(df_no_error['lib_files'])
voltages = np.array(df_no_error['voltage'])

df_ch_filter = df_no_error.drop(['ramp_files','ramp_complete','lib_files'],axis=1)
df_ch_filter.set_index('voltage',inplace=True)

# Determine the Voltages from file names
# try:
    # voltages = [int(x[-7:-4]) for x in ramp_files]   # Voltages
# except Exception as e:
    # gui.error('Voltages could not be retreived from file names. \nPlease check that the filenames are formatted correctly.')

# change voltage order, first the ref V,
for nr,i in enumerate(voltages):
    if i == V_ref:
        voltages = np.delete(voltages,nr)
        voltages = np.insert(voltages,0,V_ref)
        ramp_files = np.insert(ramp_files,0,ramp_files[nr])
        ramp_files = np.delete(ramp_files,nr+1)
        lib_files = np.insert(lib_files,0,lib_files[nr])
        lib_files = np.delete(lib_files,nr+1)


print('\n --- Evaluating  files --- \n')
# Loop over all voltages
for nr,V in enumerate(voltages):
    print('Evaluating V%d file' % V)
    # current_file_path, current_file_name = os.path.join(ramp_folder,ramp_files[nr])
    gui.task('Evaluating: %s' % ramp_files[nr])
    gui.progress('{:f}'.format(nr*100/np.size(voltages)))

    
    # Channel filter (correct PMT voltages)
    ch_filter = df_ch_filter.loc[V,:]
    ch_filter = ch_filter[ch_filter==0].index     # use to drop channels with incorrect voltage
    
    # Load ramp file
    df_ramp = load_file(os.path.join(ramp_folder,ramp_files[nr]),V)
    df_ramp[df_ramp<log_thresh_for_nan] = np.nan    # mask negative values to prevent "division by zero" through log
    df_ramp.drop(ch_filter, axis=1, inplace=True)
    

    # Drop selected channels
    drop_FSC = [x for x in df_ramp.columns if 'FSC' in x]
    df_ramp.drop(drop_FSC, axis=1, inplace = True)   
    if detector_select == 'heigth':
        drop = 'A' # Only use H files
        drop_channels = [x for x in df_ramp.columns if x[-1]==drop]
        df_ramp.drop(drop_channels, axis=1, inplace = True)

    # Evaluate Ramp file -> df_ramp_median_100npe and df_K_factor_mean
    df_gmean, df_gstd, df_CV, df_median, df_npe, df_K_factor, df_K_factor_mean, df_K_factor_info = ramp_evaluation(df_ramp)

    # Export to excel
    # Export Raw standard deviation
    if raw_SD:
        fname = os.path.join(directory, 'RAW_SD.xlsx')
        if (nr==0):
            # Initialization of excel
            sheets = ['V %s' %x for x in voltages]
            titels = ['Standard deviation for a voltage of %s V' %x for x in voltages]
            init_excel(fname, sheets, titels)
        wb = load_workbook(fname)
        writer = pd.ExcelWriter(fname, engine='openpyxl')
        writer.book = wb
        writer.sheets = dict((ws.title, ws) for ws in wb.worksheets)

        offset=4
        df_gstd.to_excel(writer, sheet_name=sheets[nr], startrow=offset)

        wb.save(fname)
        wb.close()

    # Export Raw mean
    if raw_mean:
        fname = os.path.join(directory, 'RAW_gmean.xlsx')
        if (nr==0):
            # Initialization of excel
            sheets = ['V %s' %x for x in voltages]
            titels = ['Geometric Mean position of peaks for a voltage of %s V' %x for x in voltages]
            init_excel(fname, sheets, titels)
        wb = load_workbook(fname)
        writer = pd.ExcelWriter(fname, engine='openpyxl')
        writer.book = wb
        writer.sheets = dict((ws.title, ws) for ws in wb.worksheets)

        offset=4
        df_gmean.to_excel(writer, sheet_name=sheets[nr], startrow=offset)

        wb.save(fname)
        wb.close()

    # Export Raw median
    if raw_median:
        fname = os.path.join(directory, 'RAW_median.xlsx')
        if (nr==0):
            # Initialization of excel
            sheets = ['V %s' %x for x in voltages]
            titels = ['Median position of peaks for a voltage of %s V' %x for x in voltages]
            init_excel(fname, sheets, titels)
        wb = load_workbook(fname)
        writer = pd.ExcelWriter(fname, engine='openpyxl')
        writer.book = wb
        writer.sheets = dict((ws.title, ws) for ws in wb.worksheets)

        offset=4
        df_median.to_excel(writer, sheet_name=sheets[nr], startrow=offset)

        wb.save(fname)
        wb.close()

    # Export K factor
    if K_factor:
        fname = os.path.join(directory, 'K_factor.xlsx')
        if (nr==0):
            # Initialization of excel
            sheets = ['V %s' %x for x in voltages]
            titels = ['The K_factor (npe/median) for all channels for all intensities for a voltage of %s V' %x for x in voltages]
            init_excel(fname, sheets, titels)
        wb = load_workbook(fname)
        writer = pd.ExcelWriter(fname, engine='openpyxl')
        writer.book = wb
        writer.sheets = dict((ws.title, ws) for ws in wb.worksheets)
        offset=4
        df_K_factor_info.to_excel(writer, sheet_name=sheets[nr], startrow=offset)

        offset+=8
        df_K_factor.to_excel(writer, sheet_name=sheets[nr], startrow=offset)

        # K factor plots
        # ws = wb[sheets[nr]]
        # c1 = ScatterChart()
        # c1.title = 'K Factor (npe/median) vs intensity for every channel'
        # c1.y_axis.title = 'K Factor [a.u.]'
        # c1.x_axis.title = 'Intensity [dB]'

        # c1.height = 18
        # c1.width = 30
        # Ints = Reference(ws, min_col=1,max_col=1, min_row=offset+2, max_row=offset+1+np.shape(df_K_factor)[0])

        # for col in range(4,2+np.size(df_K_factor.columns)):
            # Data = Reference(ws, min_col=col, max_col=col, min_row=offset+2, max_row = offset+1+np.shape(df_K_factor)[0])
            # title =  df_K_factor.columns[col-2]
            # s1 = Series(Data, xvalues=Ints, title=title)
            # c1.append(s1)
        # c1.y_axis.scaling.logBase = 10
        # ws.add_chart(c1, "C10")

        wb.save(fname)
        wb.close()

    if (nr==0):
        # Find npe 100 reference
        df_npe_diff = df_npe.copy() - npe_ref
        df_npe_diff = abs(df_npe_diff)
        intensity_npe_100 = df_npe_diff.abs().idxmin()
        intensity_npe_100.astype(int)

    # Initialize empty dataframes
    df_gmean_temp = intensity_npe_100.copy()
    df_gstd_temp = intensity_npe_100.copy()
    df_CV_temp = intensity_npe_100.copy()
    df_gstd_temp = intensity_npe_100.copy()
    df_median_temp = intensity_npe_100.copy()
    df_npe_temp = intensity_npe_100.copy()


    # Choose the data with the same intensity as the reference
    for channel in df_npe_temp.index:
        sel_intensity = intensity_npe_100[channel]
        df_gmean_temp.loc[channel] = df_gmean.loc[sel_intensity,channel]
        df_gstd_temp.loc[channel] = df_gstd.loc[sel_intensity,channel]
        df_CV_temp.loc[channel] = df_CV.loc[sel_intensity, channel]
        df_median_temp.loc[channel] = df_median.loc[sel_intensity,channel]
        df_npe_temp.loc[channel] = df_npe.loc[sel_intensity,channel]

    if (nr==0): # For the first voltage (ref)
        df_gmean_ref = df_gmean_temp
        df_gstd_ref = df_gstd_temp
        df_CV_ref = df_CV_temp
        df_median_ref = df_median_temp
        df_npe_ref = df_npe_temp
        df_K_factor_ref = df_K_factor_mean.copy()


    else: # For all other voltages
        df_gmean_ref = pd.concat([df_gmean_ref, df_gmean_temp], axis=1)
        df_gstd_ref =  pd.concat([df_gstd_ref, df_gstd_temp], axis=1)
        df_CV_ref =  pd.concat([df_CV_ref, df_CV_temp], axis=1)
        df_median_ref = pd.concat([df_median_ref, df_median_temp], axis=1)
        df_npe_ref = pd.concat([df_npe_ref, df_npe_temp], axis=1)
        df_K_factor_ref = pd.concat([df_K_factor_ref, df_K_factor_mean], axis=1)


    # Load Lib file for V
    df_lib = load_file(os.path.join(lib_folder,lib_files[nr]),V)
    df_lib.drop(drop_FSC, axis=1, inplace = True)   
    df_lib.drop(ch_filter, axis=1, inplace=True)
    df_lib.drop(drop_channels, axis=1, inplace = True)

    # Evaluate Lib file -> df_lib_prctile
    df_lib_prctile_temp = lib_evaluation(df_lib)
    if (nr==0):
        df_lib_prctile = df_lib_prctile_temp
    else:
        df_lib_prctile = pd.concat([df_lib_prctile, df_lib_prctile_temp], axis=1)


# Assign new column names
df_gmean_ref.columns = voltages
df_gstd_ref.columns = voltages
df_CV_ref.columns = voltages
df_median_ref.columns = voltages
df_K_factor_ref.columns = voltages
df_npe_ref.columns = voltages

df_lib_prctile.columns = voltages


# Correct npe Lib with K factors
df_npe_lib = df_lib_prctile.copy()
for V in voltages:
    df_npe_lib.loc[:,V] = df_lib_prctile.loc[:,V] * df_K_factor_ref.loc[:,V]

# Determine SNR and DNR -> df_SNR_DNR
df_SNR, df_DNR = snr_dnr(df_npe_ref, df_npe_lib, df_K_factor_ref, detector_select, resolution)
# Fix Voltage Order
# voltages.sort()
# df_gmean_ref = df_gmean_ref[voltages]
# df_gstd_ref = df_gstd_ref[voltages]
# df_CV_ref = df_CV_ref[voltages]
# df_median_ref = df_median_ref[voltages]
# df_K_factor_ref = df_K_factor_ref[voltages]
# df_npe_ref = df_npe_ref[voltages]

# df_npe_lib = df_npe_lib[voltages]
# df_lib_prctile = df_lib_prctile[voltages]

# df_SNR = df_SNR[voltages]
# df_DNR = df_DNR[voltages]
df_gmean_ref.sort_index(axis=1,inplace=True)
df_gstd_ref.sort_index(axis=1,inplace=True)
df_CV_ref.sort_index(axis=1,inplace=True)
df_median_ref.sort_index(axis=1,inplace=True)
df_K_factor_ref.sort_index(axis=1,inplace=True)
df_npe_ref.sort_index(axis=1,inplace=True)

df_npe_lib.sort_index(axis=1,inplace=True)
df_lib_prctile.sort_index(axis=1,inplace=True)

df_SNR.sort_index(axis=1,inplace=True)
df_DNR.sort_index(axis=1,inplace=True)


# Optimal results
values = np.ones([df_SNR.index.size,3])*V_ref
df_Vopt = pd.DataFrame(values, index = df_SNR.index, columns=['Vmin','Vopt','Vmax'])
#df_Vopt = optimum_voltage(df_SNR, df_DNR)
df_results = pd.concat([df_Vopt['Vopt'],df_SNR.loc[:,V_ref],df_DNR.loc[:,V_ref]],axis=1)
df_results.columns = ['V ref','SNR','DNR']

# Export evaluation results to excel
print('\n --- Export evaluation results to excel --- \n')
export_results(df_results,
               df_SNR,
               df_DNR,
               df_npe_lib,
               df_lib_prctile,
               df_npe_ref,
               df_gmean_ref,
               df_median_ref,
               df_gstd_ref,
               df_CV_ref,
               df_K_factor_ref,
               intensity_npe_100,
               data_folder)

print('\n')
print('------------------------------------')
print('DONE')
print('------------------------------------')
print('\n')
